
# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def beta_confidence_intervals(ci_X, ntrials, ci=0.95):
    """
    Compute confidence intervals of beta distributions.

    Parameters
    ----------
    ci_X : numpy.array
        Computed confidence interval estimate from `ntrials` experiments
    ntrials : int
        The number of trials that were run.
    ci : float, optional, default=0.95
        Confidence interval to report (e.g. 0.95 for 95% confidence interval)

    Returns
    -------
    Plow : float
        The lower bound of the symmetric confidence interval.
    Phigh : float
        The upper bound of the symmetric confidence interval.

    Examples
    --------

    >>> ci_X = np.random.rand(10,10)
    >>> ntrials = 100
    >>> [Plow, Phigh] = beta_confidence_intervals(ci_X, ntrials)

    """
    # Compute low and high confidence interval for symmetric CI about mean.
    ci_low = 0.5 - ci/2;
    ci_high = 0.5 + ci/2;

    # Compute for every element of ci_X.
    from scipy.stats import beta
    Plow = ci_X * 0.0;
    Phigh = ci_X * 0.0;
    for i in range(ci_X.shape[0]):
        for j in range(ci_X.shape[1]):
            Plow[i,j] = beta.ppf(ci_low, a = ci_X[i,j] * ntrials, b = (1-ci_X[i,j]) * ntrials);
            Phigh[i,j] = beta.ppf(ci_high, a = ci_X[i,j] * ntrials, b = (1-ci_X[i,j]) * ntrials);

    return [Plow, Phigh]


def empirical_confidence_interval(sample, interval=0.95):
    """
    Compute specified symmetric confidence interval for empirical sample.

    Parameters
    ----------
    sample : numpy.array
        The empirical samples.
    interval : float, optional, default=0.95
        Size of desired symmetric confidence interval (0 < interval < 1)
        e.g. 0.68 for 68% confidence interval, 0.95 for 95% confidence interval

    Returns
    -------
    low : float
        The lower bound of the symmetric confidence interval.
    high : float
        The upper bound of the symmetric confidence interval.

    Examples
    --------
    >>> sample = np.random.randn(1000)
    >>> [low, high] = empirical_confidence_interval(sample)

    >>> [low, high] = empirical_confidence_interval(sample, interval=0.65)

    >>> [low, high] = empirical_confidence_interval(sample, interval=0.99)

    """
    # Sort sample in increasing order.
    sample = np.sort(sample)

    # Determine sample size.
    N = len(sample)

    # Compute low and high indices.
    low_index = int(np.round((N-1) * (0.5 - interval/2))) + 1
    high_index = int(np.round((N-1) * (0.5 + interval/2))) + 1

    # Compute low and high.
    low = sample[low_index]
    high = sample[high_index]

    return [low, high]


def generate_latex_table(sampled_hmm, conf=0.95, dt=1, time_unit='ms', obs_name='force', obs_units='pN',
                         caption='', outfile=None):
    """
    Generate a LaTeX column-wide table showing various computed properties and uncertainties.

    Parameters
    ----------
    conf : float
        confidence interval. Use 0.68 for 1 sigma, 0.95 for 2 sigma etc.

    """
    # check input
    from bhmm.hmm.generic_sampled_hmm import SampledHMM
    from bhmm.hmm.gaussian_hmm import SampledGaussianHMM
    assert issubclass(sampled_hmm.__class__, SampledHMM), 'sampled_hmm ist not a SampledHMM'

    # confidence interval
    sampled_hmm.set_confidence(conf)
    # dt
    dt = float(dt)
    # nstates
    nstates = sampled_hmm.nstates

    table = """
\\begin{table}
    \\begin{tabular*}{\columnwidth}{@{\extracolsep{\\fill}}lcc}
        \hline
        {\\bf Property} & {\\bf Symbol} & {\\bf Value} \\\\
        \hline
            """
    # Stationary probability.
    p = sampled_hmm.stationary_distribution_mean
    p_lo, p_hi = sampled_hmm.stationary_distribution_conf
    for i in range(nstates):
        if i == 0:
            table += '\t\tEquilibrium probability '
        table += '\t\t& $\pi_{%d}$ & $%0.3f_{\:%0.3f}^{\:%0.3f}$ \\\\' % (i+1, p[i], p_lo[i], p_hi[i]) + '\n'
    table += '\t\t\hline' + '\n'

    # Transition probabilities.
    P = sampled_hmm.transition_matrix_mean
    P_lo, P_hi = sampled_hmm.transition_matrix_conf
    for i in range(nstates):
        for j in range(nstates):
            if i == 0 and j == 0:
                table += '\t\tTransition probability ($\Delta t = $%s) ' % (str(dt)+' '+time_unit)
            table += '\t\t& $T_{%d%d}$ & $%0.4f_{\:%0.4f}^{\:%0.4f}$ \\\\' % (i+1, j+1, P[i, j], P_lo[i, j], P_hi[i, j]) + '\n'
    table += '\t\t\hline' + '\n'
    table += '\t\t\hline' + '\n'

    # Transition rates via pseudogenerator.
    K = P - np.eye(sampled_hmm.nstates)
    K /= dt
    K_lo = P_lo - np.eye(sampled_hmm.nstates)
    K_lo /= dt
    K_hi = P_hi - np.eye(sampled_hmm.nstates)
    K_hi /= dt
    for i in range(nstates):
        for j in range(nstates):
            if i == 0 and j == 0:
                table += '\t\tTransition rate (%s$^{-1}$) ' % time_unit
            if i != j:
                table += '\t\t& $k_{%d%d}$ & $%2.4f_{\:%2.4f}^{\:%2.4f}$ \\\\' % (i+1, j+1, K[i, j], K_lo[i, j], K_hi[i, j]) + '\n'
    table += '\t\t\hline' + '\n'

    # State mean lifetimes.
    l = sampled_hmm.lifetimes_mean
    l *= dt
    l_lo, l_hi = sampled_hmm.lifetimes_conf
    l_lo *= dt
    l_hi *= dt
    for i in range(nstates):
        if i == 0:
            table += '\t\tState mean lifetime (%s) ' % time_unit
        table += '\t\t& $t_{%d}$ & $%.3f_{\:%.3f}^{\:%.3f}$ \\\\' % (i+1, l[i], l_lo[i], l_hi[i]) + '\n'
    table += '\t\t\hline' + '\n'

    # State relaxation timescales.
    t = sampled_hmm.timescales_mean
    t *= dt
    t_lo, t_hi = sampled_hmm.timescales_conf
    t_lo *= dt
    t_hi *= dt
    for i in range(nstates-1):
        if i == 0:
            table += '\t\tRelaxation time (%s) ' % time_unit
        table += '\t\t& $\\tau_{%d}$ & $%.3f_{\:%.3f}^{\:%.3f}$ \\\\' % (i+1, t[i], t_lo[i], t_hi[i]) + '\n'
    table += '\t\t\hline' + '\n'

    if issubclass(sampled_hmm.__class__, SampledGaussianHMM):
        table += '\t\t\hline' + '\n'

        # State mean forces.
        m = sampled_hmm.means_mean
        m_lo, m_hi = sampled_hmm.means_conf
        for i in range(nstates):
            if i == 0:
                table += '\t\tState %s mean (%s) ' % (obs_name, obs_units)
            table += '\t\t& $\mu_{%d}$ & $%.3f_{\:%.3f}^{\:%.3f}$ \\\\' % (i+1, m[i], m_lo[i], m_hi[i]) + '\n'
        table += '\t\t\hline' + '\n'

        # State force standard deviations.
        s = sampled_hmm.sigmas_mean
        s_lo, s_hi = sampled_hmm.sigmas_conf
        for i in range(nstates):
            if i == 0:
                table += '\t\tState %s std dev (%s) ' % (obs_name, obs_units)
            table += '\t\t& $s_{%d}$ & $%.3f_{\:%.3f}^{\:%.3f}$ \\\\' % (i+1, s[i], s_lo[i], s_hi[i]) + '\n'
        table += '\t\t\hline' + '\n'

    table += """
        \\hline
    \\end{tabular*}
    \\caption{{\\bf %s}}
\\end{table}
            """ % caption

    # write to file if wanted:
    if outfile is not None:
        f = open(outfile, 'w')
        f.write(table)
        f.close()

    return table



'''
Created on 07.07.2015

@author: marscher
'''
import warnings
import numbers

import numpy as np
import scipy.sparse as sp

import six
from inspect import getargspec

class NotFittedError(Exception):
    pass

def _num_samples(x):
    """Return number of samples in array-like x."""
    if hasattr(x, 'fit'):
        # Don't get num_samples from an ensembles length!
        raise TypeError('Expected sequence or array-like, got '
                        'estimator %s' % x)
    if not hasattr(x, '__len__') and not hasattr(x, 'shape'):
        if hasattr(x, '__array__'):
            x = np.asarray(x)
        else:
            raise TypeError("Expected sequence or array-like, got %s" %
                            type(x))
    if hasattr(x, 'shape'):
        if len(x.shape) == 0:
            raise TypeError("Singleton array %r cannot be considered"
                            " a valid collection." % x)
        return x.shape[0]
    else:
        return len(x)

def check_is_fitted(estimator, attributes, msg=None, all_or_any=all):
    """Perform is_fitted validation for estimator.

    Checks if the estimator is fitted by verifying the presence of
    "all_or_any" of the passed attributes and raises a NotFittedError with the
    given message.

    Parameters
    ----------
    estimator : estimator instance.
        estimator instance for which the check is performed.

    attributes : attribute name(s) given as string or a list/tuple of strings
        Eg. : ["coef_", "estimator_", ...], "coef_"

    msg : string
        The default error message is, "This %(name)s instance is not fitted
        yet. Call 'fit' with appropriate arguments before using this method."

        For custom messages if "%(name)s" is present in the message string,
        it is substituted for the estimator name.

        Eg. : "Estimator, %(name)s, must be fitted before sparsifying".

    all_or_any : callable, {all, any}, default all
        Specify whether all or any of the given attributes must exist.
    """
    if msg is None:
        msg = ("This %(name)s instance is not fitted yet. Call 'fit' with "
               "appropriate arguments before using this method.")

    if not hasattr(estimator, 'fit'):
        raise TypeError("%s is not an estimator instance." % (estimator))

    if not isinstance(attributes, (list, tuple)):
        attributes = [attributes]

    if not all_or_any([hasattr(estimator, attr) for attr in attributes]):
        raise NotFittedError(msg % {'name': type(estimator).__name__})

def _shape_repr(shape):
    """Return a platform independent reprensentation of an array shape

    Under Python 2, the `long` type introduces an 'L' suffix when using the
    default %r format for tuples of integers (typically used to store the shape
    of an array).

    Under Windows 64 bit (and Python 2), the `long` type is used by default
    in numpy shapes even when the integer dimensions are well below 32 bit.
    The platform specific type causes string messages or doctests to change
    from one platform to another which is not desirable.

    Under Python 3, there is no more `long` type so the `L` suffix is never
    introduced in string representation.

    >>> _shape_repr((1, 2))
    '(1, 2)'
    >>> one = 2 ** 64 / 2 ** 64  # force an upcast to `long` under Python 2
    >>> _shape_repr((one, 2 * one))
    '(1, 2)'
    >>> _shape_repr((1,))
    '(1,)'
    >>> _shape_repr(())
    '()'
    """
    if len(shape) == 0:
        return "()"
    joined = ", ".join("%d" % e for e in shape)
    if len(shape) == 1:
        # special notation for singleton tuples
        joined += ','
    return "(%s)" % joined


def logsumexp(arr, axis=0):
    """Computes the sum of arr assuming arr is in the log domain.

    Returns log(sum(exp(arr))) while minimizing the possibility of
    over/underflow.

    Examples
    --------

    >>> import numpy as np
    >>> from sklearn.utils.extmath import logsumexp
    >>> a = np.arange(10)
    >>> np.log(np.sum(np.exp(a)))
    9.4586297444267107
    >>> logsumexp(a)
    9.4586297444267107
    """
    arr = np.rollaxis(arr, axis)
    # Use the max to normalize, as with the log this is what accumulates
    # the less errors
    vmax = arr.max(axis=0)
    out = np.log(np.sum(np.exp(arr - vmax), axis=0))
    out += vmax
    return out

def _ensure_sparse_format(spmatrix, accept_sparse, dtype, order, copy,
                          force_all_finite):
    """Convert a sparse matrix to a given format.

    Checks the sparse format of spmatrix and converts if necessary.

    Parameters
    ----------
    spmatrix : scipy sparse matrix
        Input to validate and convert.

    accept_sparse : string, list of string or None (default=None)
        String[s] representing allowed sparse matrix formats ('csc',
        'csr', 'coo', 'dok', 'bsr', 'lil', 'dia'). None means that sparse
        matrix input will raise an error.  If the input is sparse but not in
        the allowed format, it will be converted to the first listed format.

    dtype : string, type or None (default=none)
        Data type of result. If None, the dtype of the input is preserved.

    order : 'F', 'C' or None (default=None)
        Whether an array will be forced to be fortran or c-style.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean (default=True)
        Whether to raise an error on np.inf and np.nan in X.

    Returns
    -------
    spmatrix_converted : scipy sparse matrix.
        Matrix that is ensured to have an allowed type.
    """
    if accept_sparse is None:
        raise TypeError('A sparse matrix was passed, but dense '
                        'data is required. Use X.toarray() to '
                        'convert to a dense numpy array.')
    sparse_type = spmatrix.format
    if dtype is None:
        dtype = spmatrix.dtype
    if sparse_type in accept_sparse:
        # correct type
        if dtype == spmatrix.dtype:
            # correct dtype
            if copy:
                spmatrix = spmatrix.copy()
        else:
            # convert dtype
            spmatrix = spmatrix.astype(dtype)
    else:
        # create new
        spmatrix = spmatrix.asformat(accept_sparse[0]).astype(dtype)
    if force_all_finite:
        if not hasattr(spmatrix, "data"):
            warnings.warn("Can't check %s sparse matrix for nan or inf."
                          % spmatrix.format)
        else:
            _assert_all_finite(spmatrix.data)
    if hasattr(spmatrix, "data"):
        spmatrix.data = np.array(spmatrix.data, copy=False, order=order)
    return spmatrix

def _assert_all_finite(X):
    """Like assert_all_finite, but only for ndarray."""
    X = np.asanyarray(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # everything is finite; fall back to O(n) space np.isfinite to prevent
    # false positives from overflow in sum method.
    if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
            and not np.isfinite(X).all()):
        raise ValueError("Input contains NaN, infinity"
                         " or a value too large for %r." % X.dtype)

def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance

    If seed is None, return the RandomState singleton used by np.random.
    If seed is an int, return a new RandomState instance seeded with seed.
    If seed is already a RandomState instance, return it.
    Otherwise raise ValueError.
    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)
    
def check_array(array, accept_sparse=None, dtype="numeric", order=None,
                copy=False, force_all_finite=True, ensure_2d=True,
                allow_nd=False, ensure_min_samples=1, ensure_min_features=1):
    """Input validation on an array, list, sparse matrix or similar.

    By default, the input is converted to an at least 2nd numpy array.
    If the dtype of the array is object, attempt converting to float,
    raising on failure.

    Parameters
    ----------
    array : object
        Input object to check / convert.

    accept_sparse : string, list of string or None (default=None)
        String[s] representing allowed sparse matrix formats, such as 'csc',
        'csr', etc.  None means that sparse matrix input will raise an error.
        If the input is sparse but not in the allowed format, it will be
        converted to the first listed format.

    dtype : string, type or None (default="numeric")
        Data type of result. If None, the dtype of the input is preserved.
        If "numeric", dtype is preserved unless array.dtype is object.

    order : 'F', 'C' or None (default=None)
        Whether an array will be forced to be fortran or c-style.

    copy : boolean (default=False)
        Whether a forced copy will be triggered. If copy=False, a copy might
        be triggered by a conversion.

    force_all_finite : boolean (default=True)
        Whether to raise an error on np.inf and np.nan in X.

    ensure_2d : boolean (default=True)
        Whether to make X at least 2d.

    allow_nd : boolean (default=False)
        Whether to allow X.ndim > 2.

    ensure_min_samples : int (default=1)
        Make sure that the array has a minimum number of samples in its first
        axis (rows for a 2D array). Setting to 0 disables this check.

    ensure_min_features : int (default=1)
        Make sure that the 2D array has some minimum number of features
        (columns). The default value of 1 rejects empty datasets.
        This check is only enforced when the input data has effectively 2
        dimensions or is originally 1D and ``ensure_2d`` is True. Setting to 0
        disables this check.

    Returns
    -------
    X_converted : object
        The converted and validated X.
    """
    if isinstance(accept_sparse, str):
        accept_sparse = [accept_sparse]

    # store whether originally we wanted numeric dtype
    dtype_numeric = dtype == "numeric"

    if sp.issparse(array):
        if dtype_numeric:
            dtype = None
        array = _ensure_sparse_format(array, accept_sparse, dtype, order,
                                      copy, force_all_finite)
    else:
        if ensure_2d:
            array = np.atleast_2d(array)
        if dtype_numeric:
            if hasattr(array, "dtype") and getattr(array.dtype, "kind", None) == "O":
                # if input is object, convert to float.
                dtype = np.float64
            else:
                dtype = None
        array = np.array(array, dtype=dtype, order=order, copy=copy)
        # make sure we actually converted to numeric:
        if dtype_numeric and array.dtype.kind == "O":
            array = array.astype(np.float64)
        if not allow_nd and array.ndim >= 3:
            raise ValueError("Found array with dim %d. Expected <= 2" %
                             array.ndim)
        if force_all_finite:
            _assert_all_finite(array)

    shape_repr = _shape_repr(array.shape)
    if ensure_min_samples > 0:
        n_samples = _num_samples(array)
        if n_samples < ensure_min_samples:
            raise ValueError("Found array with %d sample(s) (shape=%s) while a"
                             " minimum of %d is required."
                             % (n_samples, shape_repr, ensure_min_samples))

    if ensure_min_features > 0 and array.ndim == 2:
        n_features = array.shape[1]
        if n_features < ensure_min_features:
            raise ValueError("Found array with %d feature(s) (shape=%s) while"
                             " a minimum of %d is required."
                             % (n_features, shape_repr, ensure_min_features))
    return array



# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from abc import ABCMeta, abstractmethod


class OutputModel(object):
    """
    HMM output probability model abstract base class.

    """

    # Abstract base class.
    __metaclass__ = ABCMeta

    # implementation codes
    __IMPL_PYTHON__ = 0
    __IMPL_C__ = 1

    # implementation used
    __impl__ = __IMPL_PYTHON__

    def __init__(self, nstates, ignore_outliers=True):
        """
        Create a general output model.

        Parameters
        ----------
        nstates : int
            The number of output states.
        ignore_outliers : bool
            By outliers we mean observations that have zero probability given the
            current model. ignore_outliers=True means that outliers will be treated
            as if no observation was made, which is equivalent to making this
            observation with equal probability from any hidden state.
            ignore_outliers=True means that an Exception or in the worst case an
            unhandled crash will occur if an outlier is observed.
            If outliers have been found, the flag found_outliers will be set True

        """
        self._nstates = nstates
        self.ignore_outliers = ignore_outliers
        self.found_outliers = False

        return

    @property
    def nstates(self):
        r""" Number of hidden states """
        return self._nstates

    def set_implementation(self, impl):
        """
        Sets the implementation of this module

        Parameters
        ----------
        impl : str
            One of ["python", "c"]

        """
        if impl.lower() == 'python':
            self.__impl__ = self.__IMPL_PYTHON__
        elif impl.lower() == 'c':
            self.__impl__ = self.__IMPL_C__
        else:
            import warnings
            warnings.warn('Implementation '+impl+' is not known. Using the fallback python implementation.')
            self.__impl__ = self.__IMPL_PYTHON__

    @abstractmethod
    def sub_output_model(self, states):
        """ Returns output model on a subset of states """
        pass

    def log_p_obs(self, obs, out=None, dtype=np.float32):
        """
        Returns the element-wise logarithm of the output probabilities for an entire trajectory and all hidden states

        This is a default implementation that will take the log of p_obs(obs) and should only be used if p_obs(obs)
        is numerically stable. If there is any danger of running into numerical problems *during* the calculation of
        p_obs, this function should be overwritten in order to compute the log-probabilities directly.

        Parameters
        ----------
        obs : ndarray((T), dtype=int)
            a discrete trajectory of length T

        Return
        ------
        p_o : ndarray (T,N)
            the log probability of generating the symbol at time point t from any of the N hidden states

        """
        if out is None:
            return np.log(self.p_obs(obs))
        else:
            self.p_obs(obs, out=out, dtype=dtype)
            np.log(out, out=out)
            return out

    def _handle_outliers(self, p_o):
        """ Sets observation probabilities of outliers to uniform if ignore_outliers is set.
        Parameters
        ----------
        p_o : ndarray((T, N))
            output probabilities
        """
        if self.ignore_outliers:
            outliers = np.where(p_o.sum(axis=1)==0)[0]
            if outliers.size > 0:
                p_o[outliers, :] = 1.0
                self.found_outliers = True
        return p_o

    @abstractmethod
    def generate_observation_trajectory(self, s_t, dtype=None):
        """
        Generate synthetic observation data from a given state sequence.

        Parameters
        ----------
        s_t : numpy.array with shape (T,) of int type
            s_t[t] is the hidden state sampled at time t
        dtype : numpy.dtype, optional, default=None
            The datatype to return the resulting observations in.

        Returns
        -------
        o_t : numpy.array with shape (T,) of type dtype
            o_t[t] is the observation associated with state s_t[t]
        """
        pass



# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import unittest
from bhmm.init.gaussian import init_model_gaussian1d
import msmtools.analysis as msmana


class TestHMM(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        from bhmm import testsystems
        cls._nstates = 3
        cls._model, cls._observations, cls._states = testsystems.generate_synthetic_observations(nstates=cls._nstates,
                                                                                                 output='gaussian')

    def test_gmm(self):
        # Fit a Gaussian mixture model to obtain emission distributions and state stationary probabilities.
        from bhmm._external.sklearn import mixture
        gmm = mixture.GMM(n_components=self._nstates)
        gmm.fit(np.concatenate(self._observations)[:,None])
        assert gmm.n_components == self._nstates
        assert np.all(gmm.weights_ > 0)  # make sure we don't have empty states.

    def test_init(self):
        initial_model = init_model_gaussian1d(self._observations, self._nstates)
        assert initial_model.nstates == self._nstates
        assert msmana.is_transition_matrix(initial_model.transition_matrix)



if __name__ == "__main__":
    unittest.main()




# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import unittest
import numpy as np
import time
from bhmm import hidden
from bhmm.output_models.gaussian import GaussianOutputModel

print_speedup = False


class TestHidden(unittest.TestCase):

    def setUp(self):
        self.nexamples = 0
        self.A = []
        self.pi = []
        self.pobs = []
        self.T = []
        self.N = []
        self.logprob = []
        self.alpha = []
        self.time_alpha = []
        self.beta = []
        self.time_beta = []
        self.gamma = []
        self.time_gamma = []
        self.c = []
        self.time_c = []
        self.C = []
        self.time_C = []
        self.vpath = []
        self.time_vpath = []
        self.alpha_mem = []
        self.beta_mem = []
        self.gamma_mem = []
        self.C_mem = []

        # first toy example
        A = np.array([[0.9, 0.1],
                      [0.1, 0.9]])
        pi = np.array([0.5, 0.5])
        pobs = np.array([[0.1, 0.9],
                         [0.1, 0.9],
                         [0.1, 0.9],
                         [0.1, 0.9],
                         [0.5, 0.5],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1],
                         [0.9, 0.1]])
        self.append_example(A, pi, pobs)

        # second example
        A = np.array([[0.97, 0.02, 0.01],
                      [0.1,  0.8,  0.1],
                      [0.01, 0.02, 0.97]])
        pi = np.array([0.45, 0.1, 0.45])
        T = 10000
        means = np.array([-1.0, 0.0, 1.0])
        sigmas = np.array([0.5, 0.5, 0.5])
        gom = GaussianOutputModel(3, means=means, sigmas=sigmas)
        obs = np.random.randint(3, size=T)
        pobs = gom.p_obs(obs)
        self.append_example(A, pi, pobs)

    def append_example(self, A, pi, pobs):
        i = len(self.A)
        self.A.append(A)
        self.pi.append(pi)
        self.pobs.append(pobs)
        self.N.append(A.shape[0])
        self.T.append(pobs.shape[0])
        logprob, alpha, t = self.run_forward(i, 'python', None)
        self.logprob.append(logprob)
        self.alpha.append(alpha)
        self.time_alpha.append(t)
        beta, t = self.run_backward(i, 'python', None)
        self.beta.append(beta)
        self.time_beta.append(t)
        gamma, t = self.run_gamma(i, 'python', None)
        self.gamma.append(gamma)
        self.time_gamma.append(t)
        c, t = self.run_state_counts(i, 'python', None)
        self.c.append(c)
        self.time_c.append(t)
        C, t = self.run_transition_counts(i, 'python', None)
        self.C.append(C)
        self.time_C.append(t)
        vpath, t = self.run_viterbi(i, 'python', None)
        self.vpath.append(vpath)
        self.time_vpath.append(t)
        #
        self.alpha_mem.append(np.zeros((pobs.shape[0], A.shape[0])))
        self.beta_mem.append(np.zeros((pobs.shape[0], A.shape[0])))
        self.gamma_mem.append(np.zeros((pobs.shape[0], A.shape[0])))
        self.C_mem.append(np.zeros((A.shape[0], A.shape[0])))

        self.nexamples += 1

    def run_all(self, A, pobs, pi):
        # forward
        logprob, alpha = hidden.forward(A, pobs, pi)
        # backward
        beta = hidden.backward(A, pobs)
        # gamma
        gamma = hidden.state_probabilities(alpha, beta)
        # state counts
        T = pobs.shape[0]
        statecount = hidden.state_counts(gamma, T)
        # transition counts
        C = hidden.transition_counts(alpha, beta, A, pobs)
        # viterbi path
        vpath = hidden.viterbi(A, pobs, pi)
        # return
        return logprob, alpha, beta, gamma, statecount, C, vpath

    def run_all_mem(self, A, pobs, pi):
        T = pobs.shape[0]
        N = A.shape[0]
        alpha = np.zeros((T, N))
        beta = np.zeros((T, N))
        gamma = np.zeros((T, N))
        C = np.zeros((N, N))
        logprob, alpha = hidden.forward(A, pobs, pi, alpha_out=alpha)
        # backward
        hidden.backward(A, pobs, beta_out=beta)
        # gamma
        hidden.state_probabilities(alpha, beta, gamma_out=gamma)
        # state counts
        statecount = hidden.state_counts(gamma, T)
        # transition counts
        hidden.transition_counts(alpha, beta, A, pobs, out=self.C)
        # viterbi path
        vpath = hidden.viterbi(A, pobs, pi)
        # return
        return logprob, alpha, beta, gamma, statecount, C, vpath

    def tearDown(self):
        pass

    def run_forward(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        logprob = 0
        alpha = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            logprob, alpha = hidden.forward(self.A[i], self.pobs[i], self.pi[i], alpha_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1) / (1.0*nrep)
        return logprob, alpha, d

    def run_backward(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        beta = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            beta = hidden.backward(self.A[i], self.pobs[i], beta_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return beta, d

    def run_gamma(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        gamma = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            gamma = hidden.state_probabilities(self.alpha[i], self.beta[i], gamma_out=out)
        # compare
        time2 = time.time()
        d = (time2-time1) / (1.0*nrep)
        return gamma, d

    def run_state_counts(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        c = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            c = hidden.state_counts(self.gamma[i], self.T[i])
        # compare
        time2 = time.time()
        d = (time2-time1)/(1.0*nrep)
        return c, d

    def run_transition_counts(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        C = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            C = hidden.transition_counts(self.alpha[i], self.beta[i], self.A[i], self.pobs[i], out=out)
        # compare
        time2 = time.time()
        d = (time2-time1) / (1.0*nrep)
        return C, d

    def run_viterbi(self, i, kernel, out):
        nrep = max(1, int(10000/self.T[i]))
        vpath = None
        hidden.set_implementation(kernel)
        time1 = time.time()
        for k in range(nrep):
            vpath = hidden.viterbi(self.A[i], self.pobs[i], self.pi[i])
        # compare
        time2 = time.time()
        d = (time2-time1) / (1.0*nrep)
        return vpath, d

    def run_abs(self, call, kernel):
        """
        Reference. Just computes the time
        """
        for i in range(self.nexamples):
            res = call(i, kernel, None)
            if print_speedup:
                print('\t' + str(call.__name__) + '\t Example ' + str(i)
                      + '\t Impl = ' + str(kernel) + ' Time = ' + str(res[-1]))

    def run_comp(self, call, kernel, outs, refs, reftime):
        """
        Reference. Just computes the time
        """
        for i in range(self.nexamples):
            if outs is None:
                res = call(i, kernel, None)
            else:
                res = call(i, kernel, outs[i])
            for j in range(len(res)-1):
                myres = res[j]
                refres = refs[j][i]
                self.assertTrue(np.allclose(myres, refres))
            if outs is None:
                pkernel = kernel
            else:
                pkernel = kernel + ' mem'
            if print_speedup:
                print('\t' + str(call.__name__) + '\t Example ' + str(i) + '\t Impl = ' + pkernel
                      + ' Speedup = ' + str(reftime[i]/res[-1]))

    def test_forward_p(self):
        self.run_abs(self.run_forward, 'python')

    def test_forward_p_mem(self):
        self.run_comp(self.run_forward, 'python', self.alpha_mem, [self.logprob, self.alpha], self.time_alpha)

    def test_forward_c(self):
        self.run_comp(self.run_forward, 'c', None, [self.logprob, self.alpha], self.time_alpha)

    def test_forward_c_mem(self):
        self.run_comp(self.run_forward, 'c', self.alpha_mem, [self.logprob, self.alpha], self.time_alpha)

    def test_backward_p(self):
        self.run_abs(self.run_backward, 'python')

    def test_backward_p_mem(self):
        self.run_comp(self.run_backward, 'python', self.beta_mem, [self.beta], self.time_beta)

    def test_backward_c(self):
        self.run_comp(self.run_backward, 'c', None, [self.beta], self.time_beta)

    def test_backward_c_mem(self):
        self.run_comp(self.run_backward, 'c', self.beta_mem, [self.beta], self.time_beta)

    def test_gamma_p(self):
        self.run_abs(self.run_gamma, 'python')

    def test_gamma_p_mem(self):
        self.run_comp(self.run_gamma, 'python', self.gamma_mem, [self.gamma], self.time_gamma)

    def test_gamma_c(self):
        self.run_comp(self.run_gamma, 'c', None, [self.gamma], self.time_gamma)

    def test_gamma_c_mem(self):
        self.run_comp(self.run_gamma, 'c', self.gamma_mem, [self.gamma], self.time_gamma)

    def test_state_counts_p(self):
        self.run_abs(self.run_state_counts, 'python')

    def test_state_counts_p_mem(self):
        self.run_comp(self.run_state_counts, 'python', None, [self.c], self.time_c)

    def test_state_counts_c(self):
        self.run_comp(self.run_state_counts, 'c', None, [self.c], self.time_c)

    def test_state_counts_c_mem(self):
        self.run_comp(self.run_state_counts, 'c', None, [self.c], self.time_c)

    def test_transition_counts_p(self):
        self.run_abs(self.run_transition_counts, 'python')

    def test_transition_counts_p_mem(self):
        self.run_comp(self.run_transition_counts, 'python', self.C_mem, [self.C], self.time_C)

    def test_transition_counts_c(self):
        self.run_comp(self.run_transition_counts, 'c', None, [self.C], self.time_C)

    def test_transition_counts_c_mem(self):
        self.run_comp(self.run_transition_counts, 'c', self.C_mem, [self.C], self.time_C)

    def test_viterbi_p(self):
        self.run_abs(self.run_viterbi, 'python')

    def test_viterbi_p_mem(self):
        self.run_comp(self.run_viterbi, 'python', None, [self.vpath], self.time_vpath)

    def test_viterbi_c(self):
        self.run_comp(self.run_viterbi, 'c', None, [self.vpath], self.time_vpath)

    def test_viterbi_c_mem(self):
        self.run_comp(self.run_viterbi, 'c', None, [self.vpath], self.time_vpath)

    def test_fbtime_p_mem(self):
        for i in range(self.nexamples):
            ttot = 0.0
            logprob, alpha, t = self.run_forward(i, 'python', self.alpha_mem[i])
            ttot += t
            beta, t = self.run_backward(i, 'python', self.beta_mem[i])
            ttot += t
            gamma, t = self.run_gamma(i, 'python', self.gamma_mem[i])
            ttot += t
            c, t = self.run_state_counts(i, 'python', None)
            ttot += t
            C, t = self.run_transition_counts(i, 'python', self.C_mem[i])
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if print_speedup:
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=python mem: \t'+str(tref/ttot))

    def test_fbtime_c(self):
        for i in range(self.nexamples):
            ttot = 0.0
            logprob, alpha, t = self.run_forward(i, 'c', None)
            ttot += t
            beta, t = self.run_backward(i, 'c', None)
            ttot += t
            gamma, t = self.run_gamma(i, 'c', None)
            ttot += t
            c, t = self.run_state_counts(i, 'c', None)
            ttot += t
            C, t = self.run_transition_counts(i, 'c', None)
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if print_speedup:
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=c: \t'+str(tref/ttot))

    def test_fbtime_c_mem(self):
        for i in range(self.nexamples):
            ttot = 0.0
            logprob, alpha, t = self.run_forward(i, 'c', self.alpha_mem[i])
            ttot += t
            beta, t = self.run_backward(i, 'c', self.beta_mem[i])
            ttot += t
            gamma, t = self.run_gamma(i, 'c', self.gamma_mem[i])
            ttot += t
            c, t = self.run_state_counts(i, 'c', None)
            ttot += t
            C, t = self.run_transition_counts(i, 'c', self.C_mem[i])
            ttot += t
            tref = self.time_alpha[i] + self.time_beta[i] + self.time_gamma[i] + self.time_c[i] + self.time_C[i]
            if print_speedup:
                print ('TOTAL speedup forward-backward example '+str(i)+'\t impl=c mem: \t'+str(tref/ttot))


if __name__ == "__main__":
    unittest.main()




# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from bhmm.hmm.generic_hmm import HMM
from bhmm.util import config
from bhmm.util.statistics import confidence_interval_arr


class SampledHMM(HMM):
    """ Sampled HMM with a representative single point estimate and error estimates

    Parameters
    ----------
    estimated_hmm : :class:`HMM <generic_hmm.HMM>`
        Representative HMM estimate, e.g. a maximum likelihood estimate or mean HMM.
    sampled_hmms : list of :class:`HMM <generic_hmm.HMM>`
        Sampled HMMs
    conf : float, optional, default = 0.95
        confidence interval, e.g. 0.68 for 1 sigma or 0.95 for 2 sigma.

    """
    def __init__(self, estimated_hmm, sampled_hmms, conf=0.95):
        # call superclass constructer with estimated_hmm
        HMM.__init__(self, estimated_hmm.initial_distribution, estimated_hmm.transition_matrix,
                     estimated_hmm.output_model, lag=estimated_hmm.lag)
        # save sampled HMMs to calculate statistical moments.
        self._sampled_hmms = sampled_hmms
        self._nsamples = len(sampled_hmms)
        # save confindence interval
        self._conf = conf

    def set_confidence(self, conf):
        r""" Set confidence interval """
        self._conf = conf

    @property
    def nsamples(self):
        r""" Number of samples """
        return self._nsamples

    @property
    def sampled_hmms(self):
        r""" List of sampled HMMs """
        return self._sampled_hmms

    @property
    def confidence_interval(self):
        r""" Confidence interval used """
        return self._conf

    @property
    def initial_distribution_samples(self):
        r""" Samples of the initial distribution """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :] = self._sampled_hmms[i].stationary_distribution
        return res

    @property
    def initial_distribution_mean(self):
        r""" The mean of the initial distribution of the hidden states """
        return np.mean(self.initial_distribution_samples, axis=0)

    @property
    def initial_distribution_std(self):
        r""" The standard deviation of the initial distribution of the hidden states """
        return np.std(self.initial_distribution_samples, axis=0)

    @property
    def initial_distribution_conf(self):
        r""" The standard deviation of the initial distribution of the hidden states """
        return confidence_interval_arr(self.initial_distribution_samples, conf=self._conf)

    @property
    def stationary_distribution_samples(self):
        r""" Samples of the stationary distribution """
        if self.is_stationary:
            return self.initial_distribution_samples
        else:
            raise ValueError('HMM is not stationary')

    @property
    def stationary_distribution_mean(self):
        r""" The mean of the stationary distribution of the hidden states """
        return np.mean(self.stationary_distribution_samples, axis=0)

    @property
    def stationary_distribution_std(self):
        r""" The standard deviation of the stationary distribution of the hidden states """
        return np.std(self.stationary_distribution_samples, axis=0)

    @property
    def stationary_distribution_conf(self):
        r""" The standard deviation of the stationary distribution of the hidden states """
        return confidence_interval_arr(self.stationary_distribution_samples, conf=self._conf)

    @property
    def transition_matrix_samples(self):
        r""" Samples of the transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :, :] = self._sampled_hmms[i].transition_matrix
        return res

    @property
    def transition_matrix_mean(self):
        r""" The mean of the transition_matrix of the hidden states """
        return np.mean(self.transition_matrix_samples, axis=0)

    @property
    def transition_matrix_std(self):
        r""" The standard deviation of the transition_matrix of the hidden states """
        return np.std(self.transition_matrix_samples, axis=0)

    @property
    def transition_matrix_conf(self):
        r""" The standard deviation of the transition_matrix of the hidden states """
        return confidence_interval_arr(self.transition_matrix_samples, conf=self._conf)

    @property
    def eigenvalues_samples(self):
        r""" Samples of the eigenvalues """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :] = self._sampled_hmms[i].eigenvalues
        return res

    @property
    def eigenvalues_mean(self):
        r""" The mean of the eigenvalues of the hidden states """
        return np.mean(self.eigenvalues_samples, axis=0)

    @property
    def eigenvalues_std(self):
        r""" The standard deviation of the eigenvalues of the hidden states """
        return np.std(self.eigenvalues_samples, axis=0)

    @property
    def eigenvalues_conf(self):
        r""" The standard deviation of the eigenvalues of the hidden states """
        return confidence_interval_arr(self.eigenvalues_samples, conf=self._conf)

    @property
    def eigenvectors_left_samples(self):
        r""" Samples of the left eigenvectors of the hidden transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :, :] = self._sampled_hmms[i].eigenvectors_left
        return res

    @property
    def eigenvectors_left_mean(self):
        r""" The mean of the left eigenvectors of the hidden transition matrix """
        return np.mean(self.eigenvectors_left_samples, axis=0)

    @property
    def eigenvectors_left_std(self):
        r""" The standard deviation of the left eigenvectors of the hidden transition matrix """
        return np.std(self.eigenvectors_left_samples, axis=0)

    @property
    def eigenvectors_left_conf(self):
        r""" The standard deviation of the left eigenvectors of the hidden transition matrix """
        return confidence_interval_arr(self.eigenvectors_left_samples, conf=self._conf)

    @property
    def eigenvectors_right_samples(self):
        r""" Samples of the right eigenvectors of the hidden transition matrix """
        res = np.empty((self.nsamples, self.nstates, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :, :] = self._sampled_hmms[i].eigenvectors_right
        return res

    @property
    def eigenvectors_right_mean(self):
        r""" The mean of the right eigenvectors of the hidden transition matrix """
        return np.mean(self.eigenvectors_right_samples, axis=0)

    @property
    def eigenvectors_right_std(self):
        r""" The standard deviation of the right eigenvectors of the hidden transition matrix """
        return np.std(self.eigenvectors_right_samples, axis=0)

    @property
    def eigenvectors_right_conf(self):
        r""" The standard deviation of the right eigenvectors of the hidden transition matrix """
        return confidence_interval_arr(self.eigenvectors_right_samples, conf=self._conf)

    @property
    def timescales_samples(self):
        r""" Samples of the timescales """
        res = np.empty((self.nsamples, self.nstates-1), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :] = self._sampled_hmms[i].timescales
        return res

    @property
    def timescales_mean(self):
        r""" The mean of the timescales of the hidden states """
        return np.mean(self.timescales_samples, axis=0)

    @property
    def timescales_std(self):
        r""" The standard deviation of the timescales of the hidden states """
        return np.std(self.timescales_samples, axis=0)

    @property
    def timescales_conf(self):
        r""" The standard deviation of the timescales of the hidden states """
        return confidence_interval_arr(self.timescales_samples, conf=self._conf)

    @property
    def lifetimes_samples(self):
        r""" Samples of the timescales """
        res = np.empty((self.nsamples, self.nstates), dtype=config.dtype)
        for i in range(self.nsamples):
            res[i, :] = self._sampled_hmms[i].lifetimes
        return res

    @property
    def lifetimes_mean(self):
        r""" The mean of the lifetimes of the hidden states """
        return np.mean(self.lifetimes_samples, axis=0)

    @property
    def lifetimes_std(self):
        r""" The standard deviation of the lifetimes of the hidden states """
        return np.std(self.lifetimes_samples, axis=0)

    @property
    def lifetimes_conf(self):
        r""" The standard deviation of the lifetimes of the hidden states """
        return confidence_interval_arr(self.lifetimes_samples, conf=self._conf)




# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import sys
from bhmm.util import config


def logger(name='BHMM', pattern='%(asctime)s %(levelname)s %(name)s: %(message)s',
           date_format='%H:%M:%S', handler=logging.StreamHandler(sys.stdout)):
    """
    Retrieves the logger instance associated to the given name.

    :param name: The name of the logger instance.
    :type name: str
    :param pattern: The associated pattern.
    :type pattern: str
    :param date_format: The date format to be used in the pattern.
    :type date_format: str
    :param handler: The logging handler, by default console output.
    :type handler: FileHandler or StreamHandler or NullHandler

    :return: The logger.
    :rtype: Logger
    """
    _logger = logging.getLogger(name)
    _logger.setLevel(config.log_level())
    if not _logger.handlers:
        formatter = logging.Formatter(pattern, date_format)
        handler.setFormatter(formatter)
        handler.setLevel(config.log_level())
        _logger.addHandler(handler)
        _logger.propagate = False
    return _logger



import inspect
import warnings
import six
import numpy as np

def _pprint(params, offset=0, printer=repr):
    """Pretty print the dictionary 'params'

    Parameters
    ----------
    params: dict
        The dictionary to pretty print

    offset: int
        The offset in characters to add at the begin of each line.

    printer:
        The function to convert entries to strings, typically
        the builtin str or repr

    """
    # Do a multi-line justified repr:
    options = np.get_printoptions()
    np.set_printoptions(precision=5, threshold=64, edgeitems=2)
    params_list = list()
    this_line_length = offset
    line_sep = ',\n' + (1 + offset // 2) * ' '
    for i, (k, v) in enumerate(sorted(six.iteritems(params))):
        if type(v) is float:
            # use str for representing floating point numbers
            # this way we get consistent representation across
            # architectures and versions.
            this_repr = '%s=%s' % (k, str(v))
        else:
            # use repr of the rest
            this_repr = '%s=%s' % (k, printer(v))
        if len(this_repr) > 500:
            this_repr = this_repr[:300] + '...' + this_repr[-100:]
        if i > 0:
            if (this_line_length + len(this_repr) >= 75 or '\n' in this_repr):
                params_list.append(line_sep)
                this_line_length = len(line_sep)
            else:
                params_list.append(', ')
                this_line_length += 2
        params_list.append(this_repr)
        this_line_length += len(this_repr)

    np.set_printoptions(**options)
    lines = ''.join(params_list)
    # Strip trailing space to avoid nightmare in doctests
    lines = '\n'.join(l.rstrip(' ') for l in lines.split('\n'))
    return lines

###############################################################################
class BaseEstimator(object):
    """Base class for all estimators in scikit-learn

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        args, varargs, kw, default = inspect.getargspec(init)
        if varargs is not None:
            raise RuntimeError("scikit-learn estimators should always "
                               "specify their parameters in the signature"
                               " of their __init__ (no varargs)."
                               " %s doesn't follow this convention."
                               % (cls, ))
        # Remove 'self'
        # XXX: This is going to fail if the init is a staticmethod, but
        # who would do this?
        args.pop(0)
        args.sort()
        return args

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep: boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            warnings.simplefilter("always", DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category == DeprecationWarning:
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The former have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimisation to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)
        for key, value in six.iteritems(params):
            split = key.split('__', 1)
            if len(split) > 1:
                # nested objects case
                name, sub_name = split
                if not name in valid_params:
                    raise ValueError('Invalid parameter %s for estimator %s' %
                                     (name, self))
                sub_object = valid_params[name]
                sub_object.set_params(**{sub_name: value})
            else:
                # simple objects case
                if not key in valid_params:
                    raise ValueError('Invalid parameter %s ' 'for estimator %s'
                                     % (key, self.__class__.__name__))
                setattr(self, key, value)
        return self

    def __repr__(self):
        class_name = self.__class__.__name__
        return '%s(%s)' % (class_name, _pprint(self.get_params(deep=False),
                                               offset=len(class_name),),)







# This file is part of BHMM (Bayesian Hidden Markov Models).
#
# Copyright (c) 2016 Frank Noe (Freie Universitaet Berlin)
# and John D. Chodera (Memorial Sloan-Kettering Cancer Center, New York)
#
# BHMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
import bhmm
from os.path import abspath, join
from os import pardir


class TestMLHMM_DoubleWell(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # load observations
        testfile = abspath(join(abspath(__file__), pardir))
        testfile = join(testfile, 'data')
        testfile = join(testfile, '2well_traj_100K.dat')
        obs = np.loadtxt(testfile, dtype=int)

        # don't print
        bhmm.config.verbose = False

        # hidden states
        nstates = 2

        # run with lag 1 and 10
        cls.hmm_lag1 = bhmm.estimate_hmm([obs], nstates, lag=1, output='discrete')
        cls.hmm_lag10 = bhmm.estimate_hmm([obs], nstates, lag=10, output='discrete')

    # =============================================================================
    # Test
    # =============================================================================

    def test_output_model(self):
        from bhmm import DiscreteOutputModel
        assert isinstance(self.hmm_lag1.output_model, DiscreteOutputModel)
        assert isinstance(self.hmm_lag10.output_model, DiscreteOutputModel)

    def test_reversible(self):
        assert self.hmm_lag1.is_reversible
        assert self.hmm_lag10.is_reversible

    def test_stationary(self):
        assert not self.hmm_lag1.is_stationary
        assert not self.hmm_lag10.is_stationary

    def test_lag(self):
        assert self.hmm_lag1.lag == 1
        assert self.hmm_lag10.lag == 10

    def test_nstates(self):
        assert self.hmm_lag1.nstates == 2
        assert self.hmm_lag10.nstates == 2

    def test_transition_matrix(self):
        import msmtools.analysis as msmana
        for P in [self.hmm_lag1.transition_matrix, self.hmm_lag1.transition_matrix]:
            assert msmana.is_transition_matrix(P)
            assert msmana.is_reversible(P)

    def test_eigenvalues(self):
        for ev in [self.hmm_lag1.eigenvalues, self.hmm_lag10.eigenvalues]:
            assert len(ev) == 2
            assert np.isclose(ev[0], 1)
            assert ev[1] < 1.0

    def test_eigenvectors_left(self):
        for evec in [self.hmm_lag1.eigenvectors_left, self.hmm_lag10.eigenvectors_left]:
            assert np.array_equal(evec.shape, (2, 2))
            assert np.sign(evec[0, 0]) == np.sign(evec[0, 1])
            assert np.sign(evec[1, 0]) != np.sign(evec[1, 1])

    def test_eigenvectors_right(self):
        for evec in [self.hmm_lag1.eigenvectors_right, self.hmm_lag10.eigenvectors_right]:
            assert np.array_equal(evec.shape, (2, 2))
            assert np.isclose(evec[0, 0], evec[1, 0])
            assert np.sign(evec[0, 1]) != np.sign(evec[1, 1])

    def test_initial_distribution(self):
        for mu in [self.hmm_lag1.initial_distribution, self.hmm_lag10.initial_distribution]:
            # normalization
            assert np.isclose(mu.sum(), 1.0)
            # should be on one side
            assert np.isclose(mu[0], 1.0) or np.isclose(mu[0], 0.0)

    def test_stationary_distribution(self):
        for mu in [self.hmm_lag1.stationary_distribution, self.hmm_lag10.stationary_distribution]:
            # normalization
            assert np.isclose(mu.sum(), 1.0)
            # positivity
            assert np.all(mu > 0.0)
            # this data: approximately equal probability
            assert np.max(np.abs(mu[0]-mu[1])) < 0.05

    def test_lifetimes(self):
        for l in [self.hmm_lag1.lifetimes, self.hmm_lag10.lifetimes]:
            assert len(l) == 2
            assert np.all(l > 0.0)
        # this data: lifetimes about 680
        assert np.max(np.abs(self.hmm_lag10.lifetimes - 680)) < 20.0

    def test_timescales(self):
        for l in [self.hmm_lag1.timescales, self.hmm_lag10.timescales]:
            assert len(l) == 1
            assert np.all(l > 0.0)
        # this data: lifetimes about 680
        assert np.abs(self.hmm_lag10.timescales[0] - 340) < 20.0

class TestMLHMM_SpecialCases(unittest.TestCase):
    def test_disconnected_2state(self):
        dtrajs = [[4, 2, 0, 3, 4, 0, 1, 3, 0, 0, 3, 1, 0, 0, 1, 0, 2, 3, 2, 1, 1, 1, 2,
                   4, 0, 4, 1, 3, 1, 2, 2, 2, 3, 4, 2, 0, 1, 4, 4, 3, 3, 4, 3, 2, 2, 2,
                   2, 4, 0, 4, 2, 4, 4, 3, 3, 0, 4, 4, 3, 2, 0, 1, 1, 3, 3, 3, 0, 1, 2,
                   2, 4, 2, 1, 1, 4, 0, 3, 4, 1, 2, 4, 0, 1, 4, 2, 1, 4, 0, 4, 2, 3, 0,
                   2, 1, 0, 3, 0, 1, 3, 4],
                  [7, 9, 7, 8, 10, 6, 8, 7, 10, 9, 8, 7, 8, 6, 10, 6, 10,
                   8, 9, 6, 8, 9, 10, 7, 6, 10, 6, 9, 6, 7, 7, 9, 10, 6,
                   6, 6, 7, 7, 8, 10, 7, 10, 8, 7, 6, 10, 8, 10, 9, 6, 6,
                   8, 6, 8, 10, 10, 7, 9, 8, 7, 10, 6, 8, 6, 8, 9, 6, 6,
                   7, 7, 8, 6, 7, 10, 8, 10, 8, 10, 6, 6, 10, 10, 8, 9, 10,
                   10, 9, 8, 9, 8, 10, 7, 7, 9, 7, 10, 8, 9, 8, 10]]
        with self.assertRaises(ValueError):
            bhmm.estimate_hmm(dtrajs, 2, lag=5, output='discrete')

if __name__ == "__main__":
    unittest.main()
